/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.system.mailbox;

// タスク間メッセージングシステム


private import 	std.thread,
				std.socket;
				
private import 	pes.exception,
				pes.queue,
				pes.arraybox,
				pes.system.threadpool,
				pes.system.delay,
				pes.system.task,
				pes.system.synchronizer;

public	alias	uint			mailbox;
alias	Queue!(Mail)			MailBox;
alias	Queue!(Synchronizer)	WaiterQueue;

public const mailbox	ANONYMOUS_MAIL	= mailbox.max;

public MailResource	GlobalMail;

static this()
{
	GlobalMail = new MailResource();
}

class MailQueue
{
protected:
	MailBox[mailbox]	m_mailbox;

public:
	void create(mailbox id)
	in {
		assert((id in m_mailbox) == null);
	}
	body {
		m_mailbox[id] = new MailBox;
	}
	
	void push(mailbox id, Mail mail)
	{
		m_mailbox[id].push(mail);
	}
	
	Mail shift(mailbox id)
	{
		return m_mailbox[id].shift;
	}
	
	bool empty(mailbox id)
	{
		return m_mailbox[id].empty;
	}
	
	void rehash()
	{
		m_mailbox.rehash;
	}
	
	mailbox[] keys()
	{
		return m_mailbox.keys;
	}
	
	Object lockObject(mailbox id)
	{
		return m_mailbox[id];
	}
	
	uint length(mailbox id)
	{
		return m_mailbox[id].length;
	}
}


class MailResource
{
protected:
	MailQueue				m_mail;
	WaiterQueue[mailbox]	m_waiter;
	
public:
	this()
	{
		m_mail 		= new MailQueue;
	}
	
	void create(mailbox id)
	{
		synchronized(this) {
			m_mail.create(id);
			m_mail.rehash();
			m_waiter[id] = new WaiterQueue;
			m_waiter.rehash;
		}
	}
	
	// 待ち行列に並ぶスレッドいるとデッドロックするので注意
	void remove()
	{
		synchronized(this) {
			foreach (mailbox id; m_mail.keys) {
				while (!m_mail.empty(id)) {
					m_mail.shift(id);
				}
			}
			foreach (mailbox id; m_waiter.keys) {
				while (!m_waiter[id].empty) {
					m_waiter[id].shift.signal();
				}
			}
		}
	}
	
	void remove(mailbox id)
	{
		synchronized(this) {
			while (!m_mail.empty(id)) {
				m_mail.shift(id);
			}
			while (!m_waiter[id].empty) {
				m_waiter[id].shift.signal();
			}
		}
	}
	
	void send(mailbox id, Mail mail)
	{
		mail.to = id;
		
		synchronized(this) {
			m_mail.push(id, mail);
			
			if (! m_waiter[id].empty) {
				m_waiter[id].shift.signal();
			}
		}
		Task.yield;
	}
	
	Mail receive(mailbox id)
	{
		Mail			mail;
		Synchronizer 	sync;
		
		synchronized(this) {
			if (m_waiter[id].empty) {
				if (! m_mail.empty(id)) {
					mail 	= m_mail.shift(id);
					return mail;
				}
			}
			sync = Task.getThis.synchronizer;
			m_waiter[id].push(sync);
		}
		sync.wait();
		mail = m_mail.shift(id);
		
		return mail;
	}
	
	// 10msec単位で監視する
	// 同じメールボックスを複数のタスクが監視する場合は
	// 受信前に別タスクに取られる可能性があるので注意
	// 基本的に一人用
	bool poll(mailbox id, uint wait_msec = 0)
	{
		bool 	found = false;
		
		while (true) {
			synchronized(this) {
				if (!m_mail.empty(id)) {
					found 	= true;
					break;
				}
			}
			if (wait_msec == 0) {
				break;
			} else if (wait_msec > 10) {
				delayTask(10);
				wait_msec -= 10;
			} else {
				delayTask(wait_msec);
				wait_msec  = 0;
			}
		}
		
		return found;
	}
}

class Mail
{
protected:
	uint 		m_message_id;
	mailbox		m_from;
	mailbox		m_to;
	int			m_status;
	Object		m_data;
	
public:
	this() {}
	static Mail make(uint message_id, mailbox from, int stat, Object data)
	{
		Mail mail = new Mail;
		
		mail.m_message_id	= message_id;
		mail.m_from			= from;
		mail.m_data			= data;
		mail.m_status		= stat;
		
		return mail;
	}
	
	static Mail makeResponse(uint message_id, int stat, Object data)
	{
		return make(message_id, ANONYMOUS_MAIL, stat, data);
	}
	
	static Mail makeRequest(uint message_id, Object data = null)
	{
		return make(message_id, ANONYMOUS_MAIL, 0, data);
	}
	
	static Mail makeRequest(uint message_id, mailbox from, Object data = null)
	{
		return make(message_id, from, 0, data);
	}
	
	// pro
	
	void messageId(uint id)
	{
		m_message_id = id;
	}
	
	uint messageId()
	{
		return m_message_id;
	}
	
	void from(mailbox addr)
	{
		m_from = addr;
	}
	
	mailbox from()
	{
		return m_from;
	}
	
	void to(mailbox addr)
	{
		m_to = addr;
	}
	
	mailbox to()
	{
		return m_to;
	}
	
	void status(int stat)
	{
		m_status = stat;
	}
	
	int status()
	{
		return m_status;
	}
	
	Object data()
	{
		return m_data;
	}
	
	void data(Object data)
	{
		m_data = data;
	}
}


debug(pes_system_mailbox)
{
	import std.stdio;
	import std.c.time;
	import std.string;
	
	enum: mailbox {
		TASK1,
		TASK2,
	}
	
	class Data {
		int data;
		this (int i)
		{
			data = i;
		}
		override char[] toString()
		{
			return std.string.toString(data);
		}
	}
	
	class Task1: Task
	{
		this(mailbox id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			Mail mail;
			
			for (int i = 0; i < 2500; ++i) {
	//			msleep(500);
				mail = Mail.makeRequest(0, TASK1, new Data(i * 2));
				sendMail(TASK2, mail);
				writef("TASK1: send\n");
				
				mail = receiveMail();
				writef("TASK1: %s\n", mail.data.toString);
			}
			
			return 0;
		}
	}
	
	class Task3: Task
	{
		this(mailbox id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			Mail mail;
			
			for (int i = 0; i < 2500; ++i) {
	//			msleep(500);
				mail = Mail.makeRequest(0, TASK1, new Data(i * 2));
				sendMail(TASK2, mail);
				writef("TASK3: send\n");
				
//				mail = receiveMail();
//				writef("TASK3: %s\n", mail.data.toString);
			}
			
			return 0;
		}
	}
	
	class Task2: Task
	{
		this(mailbox id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			Mail mail;
			mailbox send_from;
			
			for (int i = 0; i < 5000; ++i) {
				writef("TASK2: receive\n");
				mail = receiveMail();
				writef("TASK2: %s\n", mail.data.toString);
				
				send_from = mail.from;
//				msleep(i % 1000);
				mail.from = TASK2;
				mail.data = new Data(i * 3);
				
				sendMail(send_from, mail);
			}
			
			return 0;
		}
	}
	
	int main()
	{
		TaskResource	resource = new TaskResource;
		resource.mailBox.create(TASK1);
		resource.mailBox.create(TASK2);
		
		Task t1 = new Task1(TASK1, resource);
		Task t3 = new Task3(TASK1, resource);
		Task t2 = new Task2(TASK2, resource);
		
		t3.start();
		t3.wait();
		
		t1.start();
		t2.start();
		
		t1.wait();
		t2.wait();
		t3.wait();
		
		return 0;
	}
}



